#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "globals.h"
#include "output.h"
#include "file_handling.h"
#include "carrier_statistics.h"
#include "occupancies.h"
#include "dispersion.h"
#include "memmory.h"

int ** histogram_ekx;
int * n_x_history;
int * e_x_history;
int n_collections;

int allocation_not_done=1;

//Main output driver
void output(char * working_folder){
	
	if (allocation_not_done){ allocate(); allocation_not_done = 0;  };

	output_n_density(working_folder,"n_density");
	output_e_density(working_folder,"e_density");
	output_Ekx(working_folder,"Ekx");
	output_rates(working_folder,"scattering_rates");
	output_current(working_folder, "current");
	output_temperatures(working_folder, "temperatures");
	output_heat(working_folder, "heat");
	output_distribution(working_folder, "energy_histogram");

	return;
}

//allocate shared arrays
void allocate(){

	int_1d_dynamic(&n_x_history, simulation.nx);
	int_1d_dynamic(&e_x_history, simulation.nx);
	int_2d_dynamic(&histogram_ekx, simulation.nx, simulation.ne);

	return;
}

//electron particle density
void output_n_density(char * working_folder,char * filename){
	int i;
	char file[n_file_chars];
	FILE* fs;
	
	create_file(file,working_folder,filename,"",1);
	fs=fopen(file,"w");

	fprintf(fs,"x n\n");
	for(i=0;i<simulation.nx;i++){
		fprintf(fs, "%g %g\n", simulation.dx*i / NM, n_x_history[i] * 1.0 / n_collections / WEIGHT);//
	}
	fclose(fs);
	return;
}

//electron energy density
void output_e_density(char * working_folder,char * filename){
	
	int i;
	char file[n_file_chars];
	FILE* fs;
	
	create_file(file,working_folder,filename,"",1);
	fs=fopen(file,"w");

	fprintf(fs,"x e\n");
	for(i=0;i<simulation.nx;i++){
		fprintf(fs, "%g %g\n", simulation.dx*i / NM, e_x_history[i] / EC / WEIGHT/n_collections);//
	}
	fclose(fs);

	return;
}

//energy momentum and position for each superparticle
void output_Ekx(char * working_folder,char * filename){
	
	int particle;
	char file[n_file_chars];
	FILE* fs;
	
	create_file(file,working_folder,filename,"",1);
	fs=fopen(file,"w");
	if (!fs){
		printf("error opening file %s\n", file);
		exit(1); }
	fprintf(fs,"i e k x\n");
	for (particle = 0; particle<simulation.np; particle++){
		if (!electrons.unused_particle[particle]){
			fprintf(fs, "%i %g %g %g\n", particle, electrons.ensemble[particle].energy / EC, electrons.ensemble[particle].momentum * NM, electrons.ensemble[particle].position / NM);
		}
	}

	fclose(fs);
	return;
}


//scattering rates
void output_rates(char * working_folder,char * filename){
	
	int i;
	int n;
	char file[n_file_chars];
	FILE* fs;
	double fp_;
	double rate_s;
	char str[10];

	create_file(file,working_folder,filename,"",1);
	fs=fopen(file,"w");

	fprintf(fs,"e ");
	for (n = 0; n < simulation.n_interactions_x2; n++){
		if (n % 2){ strcpy(str,"em_"); }
		else{ strcpy(str, "ab_"); }
		fprintf(fs,"%s%i ",str,n/2);
	}
	fprintf(fs, "sum u lambda\n");
	for(i=0;i<simulation.ne;i++){
		fprintf(fs,"%g ",i*simulation.de/EC);
		rate_s = 0.0;
		for(n=0;n<simulation.n_interactions_x2;n++){
			phonon_occupancy(&fp_, phonons.E_p[n / 2], phonons.T_pO[simulation.nx - 1][n / 2]);
			if(n % 2){
				fp_ += 1.0;
			}
			fprintf(fs, "%g ", fp_*tables.scattering[n][i]);
			rate_s += fp_*tables.scattering[n][i];
		}
		
		fprintf(fs, "%g %g %g\n",rate_s, energy_to_momentum(i*simulation.de)*H_inv2M/NM*PS, energy_to_momentum(i*simulation.de)*H_inv2M / rate_s / NM);
		
	}
	fclose(fs);
	return;
}

//cumulative current as compared to generation
void output_current(char * working_folder, char * filename){
	int i;
	char file[n_file_chars];
	FILE* fs;
	double scale = EC / WEIGHT*simulation.dx*NM;
	double t;
	create_file(file, working_folder, filename, "", 1);
	fs = fopen(file, "w");

	fprintf(fs, "t current integrated_current generation\n");
	for (i = 1; i < simulation.nt; i+=10){
		t = (i + 1)*simulation.dt;
		fprintf(fs, "%g %g %g %g\n", i*simulation.dt / PS, -outputs.current[i] * scale/t, -outputs.cumulative_current[i] * scale / t, outputs.n_generated[i] * scale / t);
	}
	fclose(fs);

	return;
}


void output_temperatures(char * working_folder, char * filename){
	int i;
	int i_i;
	char file[n_file_chars];
	FILE* fs;

	create_file(file, working_folder, filename, "", 1);
	fs = fopen(file, "w");

	fprintf(fs, "x ");
	for (i_i = 0; i_i < simulation.n_interactions; i_i++){
		fprintf(fs, "Tp fp ");
	}
	fprintf(fs, "\n");
	for (i = 0; i < simulation.nx; i++){
		fprintf(fs, "%g ", i*simulation.dx / NM);
		for (i_i = 0; i_i < simulation.n_interactions; i_i++){
			fprintf(fs, "%g %g ",phonons.T_pO[i][i_i], phonons.fp[i][i_i]);
		}
		fprintf(fs, "\n");
	}
	fclose(fs);

	return;
}

//heat absorbed and generated for each mode
void output_heat(char * working_folder, char * filename){
	int i;
	int i_i;
	char file[n_file_chars];
	FILE* fs;
	double scale = 1.0 / WEIGHT*simulation.dx*NM;
	double t;
	create_file(file, working_folder, filename, "", 1);
	fs = fopen(file, "w");

	fprintf(fs, "t heat_generated heat_absorbed\n");
	for (i = 1; i < simulation.nt; i += 10){
		fprintf(fs, "%g ", i*simulation.dt/PS);
		for (i_i = 0; i_i < simulation.n_interactions; i_i++){
			t = (i + 1)*simulation.dt;
			fprintf(fs, "%g %g ", outputs.heat_gen[i_i][i] * scale / t, outputs.heat_abs[i_i][i] * scale / t);
		}
		fprintf(fs, "\n");
	}
	fclose(fs);

	return;
}

//electron energy histogram at each position bin
void output_distribution(char * working_folder, char * filename){
	int x;
	int e;
	double * hist_sum;

	double_1d_dynamic(&hist_sum, simulation.nx);

	char file[n_file_chars];
	FILE* fs;
	create_file(file, working_folder, filename, "", 1);
	fs = fopen(file, "w");

	fprintf(fs, "energy ");
	for (x = 0; x < simulation.nx; x++){
		fprintf(fs, "%g ", x*simulation.dx/NM);
		for (e = 0; e < simulation.ne; e++){ hist_sum[x] += histogram_ekx[x][e]; }
	}
	fprintf(fs, "\n");

	for (e = 0; e < simulation.ne; e++){
		fprintf(fs, "%g ", e*simulation.de / EC);
		for (x = 0; x < simulation.nx;x++){
			fprintf(fs, "%i ", histogram_ekx[x][e]);
		}
		for (x = 0; x < simulation.nx; x++){
			fprintf(fs, "%g ", histogram_ekx[x][e] / hist_sum[x]);
		}
		fprintf(fs, "\n");
	}

	fclose(fs);
	return;
}

//gather electron energy distributions in each bin
void collect_distribution(){
	int e;
	int x;
	int particle;
	
	if (allocation_not_done){ allocate(); allocation_not_done = 0; };

	for (particle = 0; particle < simulation.np; particle++){
		if (!electrons.unused_particle[particle]){
			e = (int)(electrons.ensemble[particle].energy / simulation.de + 0.5);
			x = (int)(electrons.ensemble[particle].position / simulation.dx + 0.5);
			if (x >= 0 && x<simulation.nx && e >= 0 && e<simulation.ne){ 
				histogram_ekx[x][e]++; }
			else if(e>=simulation.ne){
				printf("Warning, some particles have energy over or very close to simulation limit -- consider increasing it!");
			}
		}
	}
	return;
}


//gather electron density
void collect_n(){
	int x;
	int * n_x;
	
	int_1d_dynamic(&n_x, simulation.nx);

	number_of_e(n_x);

	for (x = 0; x < simulation.nx; x++){
		n_x_history[x] += n_x[x];
	}
	n_collections++;
	
	return;
}

